__authors__ = ""
__copyright__ = "(c) 2014, pymal"
__license__ = "BSD License"
__contact__ = "Name Of Current Guardian of this file <email@address>"

from urllib import request
import time

import requests
try:
    import httpcache
except ImportError:
    httpcache = None
import bs4

from pymal import consts

__all__ = ['connect', 'get_next_index', 'make_list', 'check_side_content_div', 'get_content_wrapper_div']

__SESSION = requests.session()
if httpcache is not None:
    __SESSION.mount('http://', httpcache.CachingHTTPAdapter())


def url_fixer(url: str) -> str:
    url = url.encode('utf-8')
    for i in range(128, 256):
        url = url.replace(bytes([i]), '%{0:X}'.format(i).encode("utf-8"))
    return url.decode('utf-8')


def _connect(url: str, data: str=None, headers: dict or None=None,
             auth=None) -> requests.Response:
    """
    :param url: url
    :param data: data to post
    :param headers: headers to send
    :rtype : responded sock
    """
    if headers is None:
        headers = dict()

    url = url_fixer(url)

    headers['User-Agent'] = consts.USER_AGENT
    if data is not None:
        sock = __SESSION.post(url, data=data, headers=headers, auth=auth)
    else:
        sock = __SESSION.get(url, headers=headers, auth=auth)
    return sock


def connect(url: str, data: str=None, headers: dict or None=None,
            auth=None) -> str:
    """
    :param url: url
    :param data: data to post
    :param headers: headers to send
    :rtype : responded data
    """
    return _connect(url, data, headers, auth).text.strip()


def get_next_index(i: int, list_of_tags: list) -> int:
    """
    return the i after the next <br/>

    :type i: int
    :param i: an index
    :type list_of_tags: list
    :param list_of_tags: list of tags to check the i on
    :rtype: int
    """
    while i < len(list_of_tags) and list_of_tags[i].name != 'br':
        i += 1
    return i + 1


def make_list(self_list: list, i: int, list_of_tags: list) -> int:
    """
    return the index after the next <br/> and inserting all the link until it.

    :type self_list: list
    :param self_list: a list to append links to
    :type i: int
    :param i: an index
    :type list_of_tags: list
    :param list_of_tags: list of tags to check the index on
    :rtype: int
    """
    # TODO: find a way to put it out
    from pymal import Anime
    from pymal import Manga

    n_i = get_next_index(i, list_of_tags)
    for i in range(i + 1, n_i, 2):
        assert 'a' == list_of_tags[i].name, list_of_tags[i].name
        if '/anime/' in list_of_tags[i]['href']:
            self_list.append(
                Anime.Anime(int(
                    list_of_tags[i]['href'].split('/anime/')[1].split('/')[0]
                )))
        elif '/manga/' in list_of_tags[i]['href']:
            self_list.append(
                Manga.Manga(int(
                    list_of_tags[i]['href'].split('/manga/')[1].split('/')[0]
                )))
        else:
            self_list.append(
                request.urljoin(consts.HOST_NAME, list_of_tags[i]['href']))
    return n_i


def check_side_content_div(expected_text: str, div_node: bs4.element.Tag):
    span_node = div_node.span
    assert span_node is not None, div_node
    expected_text += ":"
    if ['dark_text'] != span_node['class']:
        return False
    return expected_text == span_node.text.strip()


def __get_myanimelist_div(url: str, connection_function) -> bs4.element.Tag:
    got_robot = False
    for try_number in range(consts.RETRY_NUMBER):
        time.sleep(consts.RETRY_SLEEP)
        data = connection_function(url)
        html = bs4.BeautifulSoup(data, "html5lib").html
        if html.head.find(name="meta", attrs={"name": "robots"}) is not None:
            got_robot = True
            continue
        div = html.body.find(name="div", attrs={"id": 'myanimelist'})
        if div is not None:
            return div
    assert not got_robot, "Got robot."
    assert False, "my anime list div wasnt found"


def get_content_wrapper_div(url: str, connection_function) -> bs4.element.Tag:
    myanimelist_div = __get_myanimelist_div(url, connection_function)

    # Getting content wrapper <div>
    content_wrapper_div = myanimelist_div.find(
        name="div", attrs={"id": "contentWrapper"}, recursive=False)
    assert content_wrapper_div is not None
    return content_wrapper_div


def make_start_and_end_time(start_and_end_string: str) -> tuple:
    """
    getting mal site airing / publishing format and return it as tuple(int, int)
    """
    splited = start_and_end_string.split('to')
    if len(splited) == 1:
        start_time = splited[0].strip()
        end_time = start_time
    else:
        start_time, end_time = splited
    start_time, end_time = start_time.strip(), end_time.strip()
    return make_time(start_time), make_time(end_time)


def make_time(time_string: str) -> int:
    """
    getting mal site time string format and return it as int
    """
    if '?' == time_string or consts.MALAPPINFO_NONE_TIME == time_string:
        return float('inf')
    if time_string.isdigit():
        return int(time_string)
    try:
        start_time = time.strptime(time_string, consts.SHORT_SITE_FORMAT_TIME)
    except ValueError:
        try:
            start_time = time.strptime(time_string, consts.LONG_SITE_FORMAT_TIME)
        except ValueError:
            time_string = time_string[:4] + time_string[4:].replace('00', '01')
            start_time = time.strptime(time_string, consts.MALAPPINFO_FORMAT_TIME)
    return time.mktime(start_time)


def make_counter(counter_string: str) -> int or float:
    """
    getting mal site counter string format and return it as int
    """
    if 'Unknown' == counter_string:
        return float('inf')
    return int(counter_string)



__authors__ = ""
__copyright__ = "(c) 2014, pymal"
__license__ = "BSD License"
__contact__ = "Name Of Current Guardian of this file <email@address>"

import hashlib
from xml.etree import ElementTree
from threading import Thread
from urllib import request

from pymal import consts
from pymal import decorators
from pymal import MyManga

__all__ = ['AccountMangas']


class AccountMangas(object, metaclass=decorators.SingletonFactory):
    """
    """
    __all__ = ['reading', 'completed', 'on_hold', 'dropped', 'plan_to_read',
               'reload']

    __URL = request.urljoin(consts.HOST_NAME, "malappinfo.php?u={0:s}&type=manga")

    def __init__(self, username: str, connection):
        """
        """
        self.__connection = connection
        self.__url = self.__URL.format(username)

        self.__reading = []
        self.__completed = []
        self.__on_hold = []
        self.__dropped = []
        self.__plan_to_read = []

        self.user_days_spent_watching = None

        self.map_of_lists = {
            1: self.__reading,
            2: self.__completed,
            3: self.__on_hold,
            4: self.__dropped,
            6: self.__plan_to_read,

            '1': self.__reading,
            '2': self.__completed,
            '3': self.__on_hold,
            '4': self.__dropped,
            '6': self.__plan_to_read,

            'reading': self.__reading,
            'completed': self.__completed,
            'onhold': self.__on_hold,
            'dropped': self.__dropped,
            'plantoread': self.__plan_to_read,
        }

        self._is_loaded = False

    @property
    @decorators.load
    def reading(self) -> list:
        return self.__reading

    @property
    @decorators.load
    def completed(self) -> list:
        return self.__completed

    @property
    @decorators.load
    def on_hold(self) -> list:
        return self.__on_hold

    @property
    @decorators.load
    def dropped(self) -> list:
        return self.__dropped

    @property
    @decorators.load
    def plan_to_read(self) -> list:
        return self.__plan_to_read

    def __contains__(self, item: MyManga.MyManga) -> bool:
        return any(map(lambda x: item == x, self))

    def __iter__(self):
        class AccountMangasIterator(object):

            def __init__(self, values):
                self. values = values
                self.location = 0

            def __iter__(self):
                self.location = 0
                return self

            def __next__(self):
                if self.location >= len(self.values):
                    raise StopIteration
                value = self.values[self.location]
                self.location += 1
                return value
        return AccountMangasIterator(self.reading + self.completed +
                                     self.on_hold + self.dropped +
                                     self.plan_to_read)

    def __getitem__(self, key: str or int) -> list:
        if isinstance(key, int):
            if key < len(self.reading):
                return self.reading[key]
            key -= len(self.reading)
            if key < len(self.completed):
                return self.completed[key]
            key -= len(self.completed)
            if key < len(self.on_hold):
                return self.on_hold[key]
            key -= len(self.on_hold)
            if key < len(self.dropped):
                return self.dropped[key]
            key -= len(self.dropped)
            if key < len(self.plan_to_read):
                return self.plan_to_read[key]
            raise IndexError(
                'list index out of range (the size if {0:d}'.format(len(self)))
        key = str(key)
        for mal_object in self:
            if mal_object.title == key:
                return mal_object
        KeyError("{0:s} doesn't have the anime '{1:s}'".format(
            self.__class__.__name__, key))

    def reload(self):
        resp_data = self.__connection.connect(self.__url)
        xml_tree = ElementTree.fromstring(resp_data)
        assert 'myanimelist' == xml_tree.tag, 'myanimelist == {0:s}'.format(
            xml_tree.tag)
        xml_mal_objects = list(xml_tree)
        xml_general_data = xml_mal_objects[0]
        assert 'myinfo' == xml_general_data.tag, 'myinfo == {0:s}'.format(
            xml_general_data.tag)
        l = list(xml_general_data)
        xml_user_id = l[0]
        assert 'user_id' == xml_user_id.tag, xml_user_id.tag
        assert self.__connection.user_id == int(xml_user_id.text),\
            int(xml_user_id.text)
        xml_user_name = l[1]
        assert 'user_name' == xml_user_name.tag, xml_user_name.tag
        assert self.__connection.username == xml_user_name.text.strip(),\
            xml_user_name.text.strip()
        xml_user_reading = l[2]
        assert 'user_reading' == xml_user_reading.tag, xml_user_reading.tag
        xml_user_completed = l[3]
        assert 'user_completed' == xml_user_completed.tag,\
            xml_user_completed.tag
        xml_user_onhold = l[4]
        assert 'user_onhold' == xml_user_onhold.tag, xml_user_onhold.tag
        xml_user_dropped = l[5]
        assert 'user_dropped' == xml_user_dropped.tag, xml_user_dropped.tag
        xml_user_plantoread = l[6]
        assert 'user_plantoread' == xml_user_plantoread.tag,\
            xml_user_plantoread.tag
        user_days_spent_watching = l[7]
        assert 'user_days_spent_watching' == user_days_spent_watching.tag,\
            user_days_spent_watching.tag
        self.user_days_spent_watching = float(
            user_days_spent_watching.text.strip())

        xml_mal_objects = xml_mal_objects[1:]

        self.__reading.clear()
        self.__completed.clear()
        self.__on_hold.clear()
        self.__dropped.clear()
        self.__plan_to_read.clear()

        threads = []
        for xml_mal_object in xml_mal_objects:
            if consts.DEBUG:
                self.__get_my_mal_object(xml_mal_object)
            else:
                thread = Thread(
                    target=self.__get_my_mal_object, args=(xml_mal_object, ))
                thread.start()
                threads.append(thread)

        while threads:
            threads.pop().join()

        self._is_loaded = True

    def __get_my_mal_object(self, xml_mal_object: ElementTree.Element):
        mal_object_id_xml = xml_mal_object.find('series_mangadb_id')
        assert mal_object_id_xml is not None
        mal_object_id = int(mal_object_id_xml.text)
        my_mal_object_id_xml = xml_mal_object.find('my_id')
        assert my_mal_object_id_xml is not None
        my_mal_object_id = int(my_mal_object_id_xml.text)
        mal_object = MyManga.MyManga(mal_object_id, my_mal_object_id,
                                     self.__connection,
                                     my_mal_xml=xml_mal_object)
        self.map_of_lists[mal_object.my_status].append(mal_object)

    def __len__(self):
        return sum([1 for obj in self])

    def __repr__(self):
        return "<User mangas' number is {0:d}>".format(len(self))

    def __hash__(self):
        hash_md5 = hashlib.md5()
        hash_md5.update(self.__connection.username.encode())
        hash_md5.update(self.__class__.__name__.encode())
        return int(hash_md5.hexdigest(), 16)



__authors__ = ""
__copyright__ = "(c) 2014, pymal"
__license__ = "BSD License"
__contact__ = "Name Of Current Guardian of this file <email@address>"

__all__ = ['UnauthenticatedAccountError', 'MyAnimeListApiError',
           'MyAnimeListApiUpdateError', 'MyAnimeListApiDeleteError',
           'MyAnimeListApiAddError']


class UnauthenticatedAccountError(ValueError):
    pass


class MyAnimeListApiError(RuntimeError):
    pass


class MyAnimeListApiUpdateError(MyAnimeListApiError):
    pass


class MyAnimeListApiDeleteError(MyAnimeListApiError):
    pass


class MyAnimeListApiAddError(MyAnimeListApiError):
    pass



__authors__ = ""
__copyright__ = "(c) 2014, pymal"
__license__ = "BSD License"
__contact__ = "Name Of Current Guardian of this file <email@address>"

import os
from urllib import request
from xml.etree import ElementTree

import requests
import bs4

from pymal import decorators
from pymal import Season

__all__ = ['Seasons']


class Seasons(object, metaclass=decorators.Singleton):
    """
    Lazy making of Season from online db.
    
    Attributes:
        seasons: set of Season.
    """
    __all__ = ['seasons', 'reload']

    __HOSTNAME = 'http://github.com'
    __SEASONS_URL = request.urljoin(
        __HOSTNAME, 'erengy/taiga/tree/master/data/db/season')

    def __init__(self):
        self.__seasons = set()
        self._is_loaded = False

    @property
    @decorators.load
    def seasons(self):
        return self.__seasons

    def reload(self):
        sock = requests.get(self.__SEASONS_URL)
        table = bs4.BeautifulSoup(sock.text).find(
            name="table", attrs={'class': 'files'})
        assert table is not None

        contents = table.findAll(name='td', attrs={'class': 'content'})
        assert 0 < len(contents)
        for i in contents:
            xml_url = request.urljoin(self.__HOSTNAME, i.span.a['href'])
            sock = requests.get(xml_url)
            xml_html = sock.text

            xml_body = bs4.BeautifulSoup(xml_html).find(
                name="div", attrs={"class": "code-body"}).text
            xml_element = ElementTree.fromstring(xml_body)

            anime_elements = list(xml_element)
            season_info_elements = anime_elements[0]
            assert 'info' == season_info_elements.tag
            season_name, year = season_info_elements.find('name').text.split()

            animes_ids = set()
            anime_elements = anime_elements[1:]
            for anime_element in anime_elements:
                assert 'anime' == anime_element.tag
                _, series_animedb_id_element, _, _, _ = list(anime_element)
                assert int(series_animedb_id_element.text) not in animes_ids
                animes_ids.add(int(series_animedb_id_element.text))

            season = Season.Season(season_name, year, animes_ids)
            assert season not in self.__seasons
            self.__seasons.add(season)

        self._is_loaded = True

    def __contains__(self, item) -> bool:
        return any(map(lambda x: item in x, self.seasons))

    def __repr__(self):
        return (os.linesep + '\t').join(map(str, ['<Seasons>'] +
                                            list(self.seasons)))

    def __iter__(self):
        class SeasonsIterator(object):

            def __init__(self, values):
                self.values = list(values)
                self.location = 0

            def __iter__(self):
                self.location = 0
                return self

            def __next__(self):
                if self.location >= len(self.values):
                    raise StopIteration
                value = self.values[self.location]
                self.location += 1
                return value
        return SeasonsIterator(self.seasons)

    def __len__(self) -> int:
        return sum(map(lambda x: len(x), self.seasons))



__authors__ = ""
__copyright__ = "(c) 2014, pymal"
__license__ = "BSD License"
__contact__ = "Name Of Current Guardian of this file <email@address>"

import hashlib
from urllib import request
import os

import bs4

from pymal import decorators
from pymal import consts
from pymal import global_functions
from pymal import exceptions

__all__ = ['Anime']


class Anime(object, metaclass=decorators.SingletonFactory):
    """
    """
    __all__ = ['id', 'title', 'image_url', 'english', 'synonyms', 'japanese',
               'type', 'status', 'start_time', 'end_time', 'creators',
               'genres', 'score', 'rank', 'popularity', 'synopsis',
               'adaptations', 'characters', 'sequals', 'prequel', 'spin_offs',
               'alternative_versions', 'side_stories', 'summaries', 'others',
               'parent_stories', 'alternative_settings', 'rating', 'episodes',
               'reload', 'add']
    
    __GLOBAL_MAL_URL = request.urljoin(consts.HOST_NAME, "anime/{0:d}")
    __MY_MAL_ADD_URL = request.urljoin(
        consts.HOST_NAME, 'api/animelist/add/{0:d}.xml')

    def __init__(self, mal_id: int, mal_xml=None):
        """
        """
        self.__id = mal_id
        self._is_loaded = False

        self.__mal_url = self.__GLOBAL_MAL_URL.format(self.__id)

        # Getting staff from html
        # staff from side content
        self.__title = None
        self.__image_url = None
        self.__english = ''
        self.__synonyms = None
        self.__japanese = ''
        self.__type = None
        self.__status = None
        self.__start_time = None
        self.__end_time = None
        self.__creators = dict()
        self.__genres = dict()
        self.__score = 0.0
        self.__rank = 0
        self.__popularity = 0

        self.__rating = ''
        self.__episodes = None

        # staff from main content
        # staff from row 1
        self.__synopsis = ''

        # staff from row 2
        self.__adaptations = list()
        self.__characters = list()
        self.__sequels = list()
        self.__prequels = list()
        self.__spin_offs = list()
        self.__alternative_versions = list()
        self.__side_stories = list()
        self.__summaries = list()
        self.__others = list()
        self.__parent_stories = list()
        self.__alternative_settings = list()
        self.__full_stories = list()

        self.related_str_to_list_dict = {
            'Adaptation:': self.__adaptations,
            'Character:': self.__characters,
            'Sequel:': self.__sequels,
            'Prequel:': self.__prequels,
            'Spin-off:': self.__spin_offs,
            'Alternative version:': self.__alternative_versions,
            'Side story:': self.__side_stories,
            'Summary:': self.__summaries,
            'Other:': self.__others,
            'Parent story:': self.__parent_stories,
            'Alternative setting:': self.__alternative_settings,
            'Full story:': self.__full_stories,
        }

        if mal_xml is not None:
            self.__title = mal_xml.find('series_title').text.strip()
            self.__synonyms = mal_xml.find('series_synonyms').text
            if self.__synonyms is not None:
                self.__synonyms = self.__synonyms.strip()
            # TODO: make this number to a string (or the string to a number?)
            self.__type = mal_xml.find('series_type').text.strip()
            self_status = mal_xml.find('series_status').text.strip()
            if self_status.isdigit():
                self.__status = int(self_status)
            else:
                self.__status = self_status
                print('self.__status=', self.__status)
            self.__start_time = global_functions.make_time(mal_xml.find('series_start').text.strip())
            self.__end_time = global_functions.make_time(mal_xml.find('series_end').text.strip())
            self.__image_url = mal_xml.find('series_image').text.strip()

            self.__episodes = int(mal_xml.find('series_episodes').text)

    @property
    def id(self) -> int:
        return self.__id

    @property
    def title(self) -> str:
        if self.__title is None:
            self.reload()
        return self.__title

    @property
    def image_url(self) -> str:
        if self.__image_url is None:
            self.reload()
        return self.__image_url

    @property
    @decorators.load
    def english(self) -> str:
        return self.__english

    @property
    def synonyms(self) -> str:
        if self.__synonyms is None:
            self.reload()
        return self.__synonyms

    @property
    @decorators.load
    def japanese(self) -> str:
        return self.__japanese

    @property
    def type(self) -> str:
        if self.__type is None:
            self.reload()
        return self.__type

    @property
    def status(self) -> int:
        if self.__status is None:
            self.reload()
        return self.__status

    @property
    def start_time(self) -> int:
        if self.__start_time is None:
            self.reload()
        return self.__start_time

    @property
    def end_time(self) -> int:
        if self.__end_time is None:
            self.reload()
        return self.__end_time

    @property
    @decorators.load
    def creators(self) -> dict:
        return self.__creators

    @property
    @decorators.load
    def genres(self) ->dict:
        return self.__genres

    @property
    @decorators.load
    def score(self) -> float:
        return self.__score

    @property
    @decorators.load
    def rank(self) -> int:
        return self.__rank

    @property
    @decorators.load
    def popularity(self) -> int:
        return self.__popularity

    @property
    @decorators.load
    def synopsis(self) -> str:
        return self.__synopsis

    # staff from main content
    @property
    @decorators.load
    def adaptations(self) -> list:
        return self.__adaptations

    @property
    @decorators.load
    def characters(self) -> list:
        return self.__characters

    @property
    @decorators.load
    def sequels(self) -> list:
        return self.__sequels

    @property
    @decorators.load
    def prequels(self) -> list:
        return self.__prequels

    @property
    @decorators.load
    def spin_offs(self) -> list:
        return self.__spin_offs

    @property
    @decorators.load
    def alternative_versions(self) -> list:
        return self.__alternative_versions

    @property
    @decorators.load
    def side_stories(self) -> list:
        return self.__side_stories

    @property
    @decorators.load
    def summaries(self) -> list:
        return self.__summaries

    @property
    @decorators.load
    def others(self) -> list:
        return self.__others

    @property
    @decorators.load
    def parent_stories(self) -> list:
        return self.__parent_stories

    @property
    @decorators.load
    def alternative_settings(self) -> list:
        return self.__alternative_settings

    @property
    @decorators.load
    def full_stories(self) -> list:
        return self.__full_stories

    @property
    @decorators.load
    def rating(self) -> int:
        return self.__rating

    @property
    def episodes(self) -> int:
        if self.__episodes is None:
            self.reload()
        return self.__episodes

    def reload(self):
        # Getting content wrapper <div>
        content_wrapper_div = global_functions.get_content_wrapper_div(self.__mal_url, global_functions.connect)

        # Getting title <div>
        self.__title = content_wrapper_div.h1.contents[1].strip()

        # Getting content <div>
        content_div = content_wrapper_div.find(
            name="div", attrs={"id": "content"}, recursive=False)
        if consts.DEBUG:
            assert content_div is not None

        content_table = content_div.table

        contents = content_table.tbody.tr.findAll(name="td", recursive=False)

        # Data from side content
        side_content = contents[0]
        side_contents_divs = side_content.findAll(name="div", recursive=False)

        # Getting anime image url <img>
        img_div = side_contents_divs[0]
        img_link = img_div.find(name="a")
        assert img_link is not None
        self.__image_url = img_link.img['src']

        side_contents_divs_index = 4

        # english <div>
        english_div = side_contents_divs[side_contents_divs_index]
        if global_functions.check_side_content_div('English', english_div):
            english_span, self_english = english_div.contents
            self.__english = self_english.strip()
            side_contents_divs_index += 1
        else:
            self.__english = ''

        # synonyms <div>
        synonyms_div = side_contents_divs[side_contents_divs_index]
        if global_functions.check_side_content_div('Synonyms', synonyms_div):
            synonyms_span, self_synonyms = synonyms_div.contents
            self.__synonyms = self_synonyms.strip()
            side_contents_divs_index += 1
        else:
            self.__synonyms = ''

        # japanese <div>
        japanese_div = side_contents_divs[side_contents_divs_index]
        if global_functions.check_side_content_div('Japanese', japanese_div):
            japanese_span, self_japanese = japanese_div.contents
            self.__japanese = self_japanese.strip()
            side_contents_divs_index += 1
        else:
            self.__japanese = ''

        # type <div>
        type_div = side_contents_divs[side_contents_divs_index]
        assert global_functions.check_side_content_div('Type', type_div)
        type_span, self_type = type_div.contents
        self.__type = self_type.strip()
        side_contents_divs_index += 1

        # episodes <div>
        episodes_div = side_contents_divs[side_contents_divs_index]
        assert global_functions.check_side_content_div('Episodes', episodes_div)
        episodes_span, self_episodes = episodes_div.contents
        self.__episodes = global_functions.make_counter(self_episodes.strip())

        side_contents_divs_index += 1

        # status <div>
        status_div = side_contents_divs[side_contents_divs_index]
        assert global_functions.check_side_content_div('Status', status_div)
        status_span, self.__status = status_div.contents
        self.__status = self.__status.strip()
        side_contents_divs_index += 1

        # aired <div>
        aired_div = side_contents_divs[side_contents_divs_index]
        assert global_functions.check_side_content_div('Aired', aired_div)
        aired_span, aired = aired_div.contents
        self.__start_time, self.__end_time = global_functions.make_start_and_end_time(aired)
        side_contents_divs_index += 1

        # producers <div>
        producers_div = side_contents_divs[side_contents_divs_index]
        assert global_functions.check_side_content_div('Producers', producers_div)
        for producer_link in producers_div.findAll(name='a'):
            self.__creators[producer_link.text.strip()] = producer_link['href']
        side_contents_divs_index += 1

        # genres <div>
        genres_div = side_contents_divs[side_contents_divs_index]
        assert global_functions.check_side_content_div('Genres', genres_div)
        for genre_link in genres_div.findAll(name='a'):
            self.__genres[genre_link.text.strip()] = genre_link['href']
        side_contents_divs_index += 1

        side_contents_divs_index += 1

        # rating <div>
        rating_div = side_contents_divs[side_contents_divs_index]
        assert global_functions.check_side_content_div('Rating', rating_div)
        rating_span, self.__rating = rating_div.contents
        self.__rating = self.__rating.strip()
        side_contents_divs_index += 1

        # score <div>
        score_div = side_contents_divs[side_contents_divs_index]
        assert global_functions.check_side_content_div('Score', score_div)
        score_span, self_score = score_div.contents[:2]
        self.__score = float(self_score)
        side_contents_divs_index += 1

        # rank <div>
        rank_div = side_contents_divs[side_contents_divs_index]
        assert global_functions.check_side_content_div('Ranked', rank_div)
        rank_span, self_rank = rank_div.contents[:2]
        self_rank = self_rank.strip()
        assert self_rank.startswith("#")
        self.__rank = int(self_rank[1:])
        side_contents_divs_index += 1

        # popularity <div>
        popularity_div = side_contents_divs[side_contents_divs_index]
        assert global_functions.check_side_content_div('Popularity', popularity_div)
        popularity_span, self_popularity = popularity_div.contents[:2]
        self_popularity = self_popularity.strip()
        assert self_popularity.startswith("#")
        self.__popularity = int(self_popularity[1:])

        # Data from main content
        main_content = contents[1]
        main_content_inner_divs = main_content.findAll(
            name='div', recursive=False)
        if consts.DEBUG:
            assert 2 == len(main_content_inner_divs), \
                "Got len(main_content_inner_divs) == {0:d}".format(
                    len(main_content_inner_divs))
        main_content_datas = main_content_inner_divs[
            1].table.tbody.findAll(name="tr", recursive=False)

        synopsis_cell = main_content_datas[0]
        main_content_other_data = main_content_datas[1]

        # Getting synopsis
        synopsis_cell = synopsis_cell.td
        synopsis_cell_contents = synopsis_cell.contents
        if consts.DEBUG:
            assert 'Synopsis' == synopsis_cell.h2.text.strip(
            ), synopsis_cell.h2.text.strip()
        self.__synopsis = os.linesep.join([
            synopsis_cell_content.strip()
            for synopsis_cell_content in synopsis_cell_contents[1:-1]
            if isinstance(synopsis_cell_content, bs4.element.NavigableString)
        ])

        # Getting other data
        main_content_other_data = main_content_other_data.td
        other_data_kids = [i for i in main_content_other_data.children]

        # Getting all the data under 'Related Anime'
        index = 0
        index = global_functions.get_next_index(index, other_data_kids)
        if 'h2' == other_data_kids[index].name and\
           'Related Anime' == other_data_kids[index].text.strip():
            index += 1
            while other_data_kids[index + 1].name != 'br':
                index = global_functions.make_list(
                    self.related_str_to_list_dict[
                        other_data_kids[index].strip()],
                    index, other_data_kids)
        else:
            index -= 2
        next_index = global_functions.get_next_index(index, other_data_kids)

        if consts.DEBUG:
            assert next_index - \
                index == 2, "{0:d} - {1:d}".format(next_index, index)
            index = next_index + 1

            # Getting all the data under 'Characters & Voice Actors'
            assert 'h2' == other_data_kids[index].name, 'h2 == {0:s}'.format(
                other_data_kids[index].name)
            assert 'Characters & Voice Actors' == other_data_kids[index].contents[-1],\
                other_data_kids[index].contents[-1]

        self._is_loaded = True

    @property
    def MY_MAL_XML_TEMPLATE(self) -> str:
        return """<?xml version="1.0" encoding="UTF-8"?>
<entry>
    <episode>{0:d}</episode>
    <status>{1:d}</status>
    <score>{2:d}</score>
    <downloaded_episodes>{3:d}</downloaded_episodes>
    <storage_type>{4:d}</storage_type>
    <storage_value>{5:f}</storage_value>
    <times_rewatched>{6:d}</times_rewatched>
    <rewatch_value>{7:d}</rewatch_value>
    <date_start>{8:s}</date_start>
    <date_finish>{9:s}</date_finish>
    <priority>{10:d}</priority>
    <enable_discussion>{11:d}</enable_discussion>
    <enable_rewatching>{12:d}</enable_rewatching>
    <comments>{13:s}</comments>
    <fansub_group>{14:s}</fansub_group>
    <tags>{15:s}</tags>
</entry>"""

    def add(self, account):
        """
        """
        data = self.MY_MAL_XML_TEMPLATE.format(
            0, 6, 0, 0, 0, 0, 0, 0, consts.MALAPI_NONE_TIME,
            consts.MALAPI_NONE_TIME, 0, False, False, '', '', ''
        )
        xml = ''.join(map(lambda x: x.strip(), data.splitlines()))
        delete_url = self.__MY_MAL_ADD_URL.format(self.id)
        ret = account.auth_connect(
            delete_url,
            data='data=' + xml,
            headers={'Content-Type': 'application/x-www-form-urlencoded'}
        )
        try:
            html_obj = bs4.BeautifulSoup(ret)
            assert html_obj is not None

            head_obj = html_obj.head
            assert head_obj is not None

            title_obj = head_obj.title
            assert title_obj is not None

            data = title_obj.text
            assert data is not None

            my_id, string = data.split()
            assert my_id.isdigit()
            assert string == 'Created'
        except Exception:
            raise exceptions.MyAnimeListApiAddError(ret)

        from pymal import MyAnime
        return MyAnime.MyAnime(self, my_id, account)

    def __eq__(self, other):
        if isinstance(other, Anime):
            return self.id == other.id
        elif isinstance(other, int):
            return self.id == other
        elif isinstance(other, str) and other.isdigit():
            return self.id == int(other)
        elif hasattr(other, 'id'):
            return self.id == other.id
        return False

    def __hash__(self):
        hash_md5 = hashlib.md5()
        hash_md5.update(str(self.id).encode())
        hash_md5.update(b'Anime')
        return int(hash_md5.hexdigest(), 16)

    def __repr__(self):
        title = '' if self.__title is None else ' ' + self.__title
        return "<{0:s}{1:s} id={2:d}>".format(self.__class__.__name__, title,
                                              self.__id)



__authors__ = ""
__copyright__ = "(c) 2014, pymal"
__license__ = "BSD License"
__contact__ = "Name Of Current Guardian of this file <email@address>"

import hashlib
from urllib import request
import time

from pymal import consts
from pymal import decorators
from pymal import Manga
from pymal import global_functions
from pymal import exceptions

__all__ = ['MyManga']


class MyManga(object, metaclass=decorators.SingletonFactory):
    """
    Saves an account data about manga.
    
    Attributes:
        my_enable_discussion - boolean
        my_id - int
        my_status - int.  #TODO: put the dictanary here.
        my_score - int.
        my_start_date - string as mmddyyyy.
        my_end_date - string as mmddyyyy.
        my_priority - int.
        my_storage_type - int.  #TODO: put the dictnary here.
        my_is_rereading - boolean.
        my_completed_chapters - int.
        my_completed_volumes - int.
        my_downloaded_chapters - int.
        my_times_reread - int.
        my_reread_value - int.
        my_tags - string.
        my_comments - string
        my_fan_sub_groups - string.
    """
    __all__ = ['my_enable_discussion', 'my_id', 'my_status', 'my_score',
               'my_start_date', 'my_end_date', 'my_priority',
               'my_storage_type', 'my_is_rereading',
               'my_completed_chapters', 'my_completed_volumes',
               'my_downloaded_chapters', 'my_times_reread', 'my_reread_value',
               'my_tags', 'my_comments', 'my_fan_sub_groups', 'my_reload',
               'update', 'delete']

    __TAG_SEPARATOR = ';'
    __MY_MAL_URL = request.urljoin(
        consts.HOST_NAME, 'panel.php?go=editmanga&id={0:d}')
    __MY_MAL_DELETE_URL = request.urljoin(
        consts.HOST_NAME, 'api/mangalist/delete/{0:d}.xml')
    __MY_MAL_UPDATE_URL = request.urljoin(
        consts.HOST_NAME, 'api/mangalist/update/{0:d}.xml')

    def __init__(self, mal_id: int or Manga.Manga, my_mal_id, account,
                 my_mal_xml: None=None):
        """
        """
        if isinstance(mal_id, Manga.Manga):
            self.obj = mal_id
        else:
            self.obj = Manga.Manga(mal_id, mal_xml=my_mal_xml)

        self.__my_mal_url = self.__MY_MAL_URL.format(my_mal_id)

        self._is_my_loaded = False
        self._account = account

        self.__my_mal_id = my_mal_id
        self.__my_status = None
        self.my_enable_discussion = False
        self.__my_score = None
        self.__my_start_date = None
        self.__my_end_date = None
        self.__my_priority = 0
        self.__my_storage_type = 0
        self.__my_comments = ''
        self.__my_fan_sub_groups = ''
        self.__my_tags = None
        self.__my_retail_volumes = 0

        self.__my_is_rereading = None
        self.__my_completed_chapters = None
        self.__my_completed_volumes = None
        self.__my_downloaded_chapters = 0
        self.__my_times_reread = 0
        self.__my_reread_value = None

        if my_mal_xml is not None:
            self.__my_id = int(my_mal_xml.find('my_id').text)
            self.__my_status = int(my_mal_xml.find('my_status').text)
            if my_mal_xml.find('my_rereadingg').text is not None:
                self.__my_is_rereading = bool(
                    int(my_mal_xml.find('my_rereadingg').text))
            else:
                self.__my_is_rereading = False
            self.__my_completed_episodes = int(
                my_mal_xml.find('my_read_chapters').text.strip())
            self.__my_completed_volumes = int(
                my_mal_xml.find('my_read_volumes').text.strip())
            self.__my_score = int(my_mal_xml.find('my_score').text)
            my_start_date = my_mal_xml.find('my_start_date').text.strip()
            if my_start_date == consts.MALAPPINFO_NONE_TIME:
                self.__my_start_date = consts.MALAPI_NONE_TIME
            else:
                my_start_date = time.strptime(
                    my_start_date, consts.MALAPPINFO_FORMAT_TIME)
                self.__my_start_date = time.strftime(
                    consts.MALAPI_FORMAT_TIME, my_start_date)
            my_end_date = my_mal_xml.find('my_finish_date').text.strip()
            if my_end_date == consts.MALAPPINFO_NONE_TIME:
                self.__my_end_date = consts.MALAPI_NONE_TIME
            else:
                my_end_date = time.strptime(
                    my_end_date, consts.MALAPPINFO_FORMAT_TIME)
                self.__my_end_date = time.strftime(
                    consts.MALAPI_FORMAT_TIME, my_end_date)
            self.__my_reread_value = int(
                my_mal_xml.find('my_rereading_chap').text.strip())
            my_tags_xml = my_mal_xml.find('my_tags')
            if my_tags_xml.text is None:
                self.__my_tags = ''
            else:
                self.__my_tags = my_tags_xml.text.strip().split(
                    self.__TAG_SEPARATOR)

    @property
    def my_id(self):
        return self.__my_mal_id

    @property
    def my_status(self):
        if self.__my_status is None:
            self.my_reload()
        return self.__my_status

    @my_status.setter
    def my_status(self, value: int):
        if not (1 <= value <= 6):
            raise RuntimeError("value of my_statue can be 1 to 6")
        self.__my_status = value

    @property
    def my_score(self):
        if self.__my_score is None:
            self.my_reload()
        return self.__my_score

    @my_score.setter
    def my_score(self, value: int):
        if not (0 <= value <= 10):
            raise RuntimeError("value of my_score can be 0 to 10")
        self.__my_score = value

    @property
    def my_start_date(self):
        if self.__my_start_date is None:
            self.my_reload()
        return self.__my_start_date

    @my_start_date.setter
    def my_start_date(self, value: str):
        time.strptime(value, consts.MALAPI_FORMAT_TIME)
        self.__my_start_date = value

    @property
    def my_end_date(self):
        if self.__my_end_date is None:
            self.my_reload()
        return self.__my_end_date

    @my_end_date.setter
    def my_end_date(self, value: str):
        time.strptime(value, consts.MALAPI_FORMAT_TIME)
        self.__my_end_date = value

    @property
    @decorators.my_load
    def my_priority(self):
        return self.__my_priority

    @my_priority.setter
    def my_priority(self, value: int):
        if not (0 <= value <= 3):
            raise RuntimeError("value of my_priority can be 0 to 3")
        self.__my_priority = value

    @property
    @decorators.my_load
    def my_storage_type(self):
        return self.__my_storage_type

    @my_storage_type.setter
    def my_storage_type(self, value: int):
        if not (0 <= value <= 7):
            raise RuntimeError("value of my_storage_type can be 0 to 7")
        self.__my_storage_type = value

    @property
    @decorators.my_load
    def my_storage_value(self):
        return self.__my_storage_value

    @my_storage_type.setter
    def my_storage_value(self, value: float):
        int(value)
        self.__my_storage_value = value

    @property
    def my_is_rereading(self):
        if self.__my_is_rereading is None:
            self.my_reload()
        return self.__my_is_rereading

    @my_is_rereading.setter
    def my_is_rereading(self, value: bool):
        self.__my_is_rereading = value

    @property
    def my_completed_chapters(self):
        if self.__my_completed_chapters is None:
            self.my_reload()
        return self.__my_completed_chapters

    @my_completed_chapters.setter
    def my_completed_chapters(self, value: int):
        if not (0 <= value <= self.chapters):
            raise RuntimeError("value of my_completed_episodes can be 0 to self.chapters")
        self.__my_completed_chapters = value

    @property
    def my_completed_volumes(self):
        if self.__my_completed_volumes is None:
            self.my_reload()
        return self.__my_completed_volumes

    @my_completed_volumes.setter
    def my_completed_volumes(self, value: int):
        if not (0 <= value <= self.volumes):
            raise RuntimeError("value of my_completed_volumes can be 0 to self.volumes")
        self.__my_completed_volumes = value

    @property
    @decorators.my_load
    def my_downloaded_chapters(self):
        return self.__my_downloaded_chapters

    @my_downloaded_chapters.setter
    def my_downloaded_chapters(self, value: int):
        if not (0 <= value <= self.chapters):
            raise RuntimeError("value of my_downloaded_chapters can be 0 to self.episodes")
        self.__my_downloaded_chapters = value

    @property
    def my_times_reread(self):
        if self.__my_times_reread is None:
            self.my_reload()
        return self.__my_times_reread

    @my_times_reread.setter
    def my_times_reread(self, value: int):
        if not (0 <= value):
            raise RuntimeError("value of my_times_reread can be 0 or more")
        self.__my_times_reread = value

    @property
    def my_reread_value(self):
        if self.__my_reread_value is None:
            self.my_reload()
        return self.__my_reread_value

    @my_reread_value.setter
    def my_reread_value(self, value: int):
        if not (0 <= value <= 5):
            raise RuntimeError("value of my_reread_value can be 0 to 5")
        self.__my_reread_value = value

    @property
    def my_tags(self):
        if self.__my_tags is None:
            self.my_reload()
        return self.__my_tags

    @property
    @decorators.my_load
    def my_comments(self):
        return self.__my_comments

    @property
    @decorators.my_load
    def my_fan_sub_groups(self):
        return self.__my_fan_sub_groups

    @property
    @decorators.my_load
    def my_retail_volumes(self):
        return self.__my_retail_volumes

    def my_reload(self):
        # Getting content wrapper <div>
        content_wrapper_div = global_functions.get_content_wrapper_div(
            self.__my_mal_url, self._account.auth_connect)

        bas_result = content_wrapper_div.find(name='div',
                                              attrs={'class': 'badresult'})
        assert bas_result is None

        # Getting content <td>
        content_div = content_wrapper_div.find(
            name="div", attrs={"id": "content"}, recursive=False)
        assert content_div is not None

        content_td = content_div.table.tr.td
        assert content_td is not None

        # Getting content rows <tr>
        content_form = content_td.find(name="form", attrs={'id': "mangaForm"})
        assert 'mangaForm' == content_form['id'], content_form['id']
        content_rows = content_form.table.tbody.findAll(
            name="tr", recursive=False)

        contents_divs_index = 2

        # Getting my_status
        status_select = content_rows[contents_divs_index].find(
            name="select", attrs={"id": "status", "name": "status"})
        assert status_select is not None
        # TODO: make this look better
        status_selected_options = [
            x
            for x in status_select.findAll(name="option")
            if 'selected' in x.attrs
        ]
        assert 1 == len(status_selected_options)
        self.__my_status = int(status_selected_options[0]['value'])

        is_reread_node = content_rows[contents_divs_index].find(
            name="input", attrs={"id": "rereadingBox"})
        assert is_reread_node is not None
        self.__my_is_rereading = bool(is_reread_node['value'])
        contents_divs_index += 1

        # Getting read volumes
        read_input = content_rows[contents_divs_index].\
            find(name="input", attrs={"id": "vol_read",
                                      "name": "vol_read"})
        assert read_input is not None
        self.__my_completed_volumes = int(read_input['value'])
        contents_divs_index += 1

        # Getting read chapters
        read_input = content_rows[contents_divs_index].\
            find(name="input", attrs={"id": "chap_read",
                                      "name": "chap_read"})
        assert read_input is not None
        self.__my_completed_chapters = int(read_input['value'])
        contents_divs_index += 1

        # Getting my_score
        score_select = content_rows[contents_divs_index].find(
            name="select", attrs={"name": "score"})
        assert score_select is not None
        score_selected_option = score_select.find(
            name="option", attrs={"selected": ""})
        assert score_selected_option is not None
        self.__my_score = int(float(score_selected_option['value']))
        contents_divs_index += 1

        # Getting my_tags...
        tag_content = content_rows[contents_divs_index]
        tag_textarea = tag_content.find(
            name="textarea", attrs={"name": "tags"})
        self.__my_tags = tag_textarea.text
        contents_divs_index += 1

        # Getting start date
        start_month_date_node = content_rows[contents_divs_index].find(
            name="select", attrs={"name": "startMonth"})
        assert start_month_date_node is not None
        start_month_date = start_month_date_node.find(
            name="option", attrs={"selected": ""})

        start_day_date_node = content_rows[contents_divs_index].find(
            name="select", attrs={"name": "startDay"})
        assert start_day_date_node is not None
        start_day_date = start_day_date_node.find(
            name="option", attrs={"selected": ""})

        start_year_date_node = content_rows[contents_divs_index].find(
            name="select", attrs={"name": "startYear"})
        assert start_year_date_node is not None
        start_year_date = start_year_date_node.find(
            name="option", attrs={"selected": ""})

        start_month_date = str(start_month_date['value']).zfill(2)
        start_day_date = str(start_day_date['value']).zfill(2)
        start_year_date = str(start_year_date['value']).zfill(2)
        self.__my_start_date = start_month_date + \
            start_day_date + start_year_date
        contents_divs_index += 1

        # Getting end date
        end_month_date_node = content_rows[contents_divs_index].find(
            name="select", attrs={"name": "endMonth"})
        assert end_month_date_node is not None
        end_month_date = end_month_date_node.find(
            name="option", attrs={"selected": ""})

        end_day_date_node = content_rows[contents_divs_index].find(
            name="select", attrs={"name": "endDay"})
        assert end_day_date_node is not None
        end_day_date = end_day_date_node.find(
            name="option", attrs={"selected": ""})

        end_year_date_node = content_rows[contents_divs_index].find(
            name="select", attrs={"name": "endYear"})
        assert end_year_date_node is not None
        end_year_date = end_year_date_node.find(
            name="option", attrs={"selected": ""})

        end_month_date = str(end_month_date['value']).zfill(2)
        end_day_date = str(end_day_date['value']).zfill(2)
        end_year_date = str(end_year_date['value']).zfill(2)
        self.__my_end_date = end_month_date + end_day_date + end_year_date
        contents_divs_index += 1

        # Getting priority
        priority_node = content_rows[contents_divs_index].find(
            name="select", attrs={"name": "priority"})
        assert priority_node is not None
        selected_priority_node = priority_node.find(
            name="option", attrs={"selected": ""})
        assert selected_priority_node is not None
        self.__my_priority = int(selected_priority_node['value'])
        contents_divs_index += 1

        # Getting storage
        storage_type_node = content_rows[contents_divs_index].find(
            name="select", attrs={"id": "storageSel"})
        assert storage_type_node is not None
        selected_storage_type_node = storage_type_node.find(
            name="option", attrs={"selected": ""})
        if selected_storage_type_node is None:
            self.__my_storage_type = 0
        else:
            self.__my_storage_type = int(selected_storage_type_node['value'])
        contents_divs_index += 1

        # Getting downloaded episodes
        downloaded_chapters_node = content_rows[contents_divs_index].\
            find(name="input", attrs={'id': "dChap",
                                      'name': 'downloaded_chapters'})
        assert downloaded_chapters_node is not None
        self.__my_downloaded_chapters == int(downloaded_chapters_node['value'])
        contents_divs_index += 1

        # Getting time reread
        times_reread_node = content_rows[contents_divs_index].find(
            name="input", attrs={'name': 'times_read'})
        self.__my_times_reread == int(times_reread_node['value'])
        assert times_reread_node is not None
        contents_divs_index += 1

        # Getting reread value
        reread_value_node = content_rows[contents_divs_index].find(
            name="select", attrs={'name': 'reread_value'})
        assert reread_value_node is not None
        reread_value_option = reread_value_node.find(
            name='option', attrs={'selected': ''})
        if reread_value_option is None:
            self.__my_reread_value = 0
        else:
            self.__my_reread_value = int(reread_value_option['value'])
        contents_divs_index += 1

        # Getting comments
        comment_content = content_rows[contents_divs_index]
        comment_textarea = comment_content.find(
            name="textarea", attrs={"name": "comments"})
        self.__my_comments = comment_textarea.text
        contents_divs_index += 1

        # Getting discuss flag
        discuss_node = content_rows[contents_divs_index].find(
            name='input', attrs={"name": "discuss"})
        assert discuss_node is not None
        self._is_my_loaded = True

    def to_xml(self):
        data = self.MY_MAL_XML_TEMPLATE.format(
            self.my_completed_chapters,
            self.my_completed_volumes,
            self.my_status,
            self.my_score,
            self.my_downloaded_chapters,
            self.my_times_reread,
            self.my_reread_value,
            self.my_start_date,
            self.my_end_date,
            self.my_priority,
            self.my_is_rereading,
            self.my_enable_discussion,
            self.my_comments,
            self.my_fan_sub_groups,
            self.my_tags,
            self.my_retail_volumes
        )
        return data

    def add(self, account):
        if account == self._account:
            return self
        return self.obj.add(account)

    def update(self):
        """
        """
        xml = ''.join(map(lambda x: x.strip(), self.to_xml().splitlines()))
        update_url = self.__MY_MAL_UPDATE_URL.format(self.id)
        ret = self._account.auth_connect(
            update_url,
            data='data=' + xml,
            headers={'Content-Type': 'application/x-www-form-urlencoded'}
        )
        if ret != 'Updated':
            raise exceptions.MyAnimeListApiUpdateError(ret)

    def delete(self):
        """
        """
        xml = ''.join(map(lambda x: x.strip(), self.to_xml().splitlines()))
        delete_url = self.__MY_MAL_DELETE_URL.format(self.id)
        ret = self._account.auth_connect(
            delete_url,
            data='data=' + xml,
            headers={'Content-Type': 'application/x-www-form-urlencoded'}
        )
        if ret != 'Deleted':
            raise exceptions.MyAnimeListApiDeleteError(ret)

    def increase(self) -> bool:
        if self.is_completed:
            return False
        self.my_completed_chapters += 1
        return True

    def increase_volume(self) -> bool:
        if self.__my_downloaded_volumes >= self.obj.volumes:
            return False
        self.my_completed_volumes += 1
        return True

    def increase_downloaded(self) -> bool:
        if self.is_completed:
            return False
        self.my_downloaded_chapters += 1
        return True

    @property
    def is_completed(self) -> bool:
        return self.my_completed_chapters >= self.obj.chapters

    def set_completed(self) -> bool:
        if self.obj.chapters == float('inf'):
            return False
        self.my_completed_chapters = self.obj.chapters
        if self.obj.volumes != float('inf'):
            self.my_completed_volumes = self.obj.volumes
        return True

    def set_completed_download(self) -> bool:
        if self.obj.chapters == float('inf'):
            return False
        self.my_downloaded_chapters = self.obj.chapters
        return True

    def __getattr__(self, name):
        return getattr(self.obj, name)

    def __dir__(self):
        return list(set(dir(type(self)) + list(self.__dict__.keys()) + dir(self.obj)))

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return hash(other) == hash(self)
        if isinstance(other, self.obj.__class__):
            return hash(other) == hash(self.obj)
        return False

    def __hash__(self):
        hash_md5 = hashlib.md5()
        hash_md5.update(str(self.id).encode())
        hash_md5.update(str(hash(self._account)).encode())
        hash_md5.update(b'MyManga')
        return int(hash_md5.hexdigest(), 16)

    def __repr__(self):
        title = " '{0:s}'".format(self.title) if self.obj._is_loaded else ''
        return "<{0:s}{1:s} of account '{2:s}' id={3:d}>".format(
            self.__class__.__name__, title, self._account.username, self.id)



__authors__ = ""
__copyright__ = "(c) 2014, pymal"
__license__ = "BSD License"
__contact__ = "Name Of Current Guardian of this file <email@address>"

import hashlib
import time

from pymal import decorators
from pymal import Anime

__all__ = ['Season']


class Season(object, metaclass=decorators.SingletonFactory):
    """
    Lazy load of season data.
    
    Attributes:
        animes - a set of animes.
        year - the season year.
        season_name - The season name.
          Can be 'Winter', 'Spring', 'Summer' or 'Fall'.
    """
    __all__ = ['animes', 'reload']

    __SEAONS_NAME_TO_START_MONTH = {
        'Winter': 1,
        'Spring': 4,
        'Summer': 7,
        'Fall': 10
    }

    def __init__(self, season_name: str, year: int or str, animes_ids: set):
        """
        """
        self.season_name = season_name.title()
        self.year = int(year)
        self.__animes_ids = animes_ids
        self._is_loaded = False
        self.__animes = set()

        month = str(self.__SEAONS_NAME_TO_START_MONTH[self.season_name])
        start_time_string = str(year) + ' ' + month
        self.start_time = time.strptime(start_time_string, '%Y %m')

    @property
    @decorators.load
    def animes(self):
        return self.__animes

    def reload(self):
        self.__animes = set(map(Anime.Anime, self.__animes_ids))
        self._is_loaded = True
        assert len(self.__animes) == len(self.__animes_ids)

    def __iter__(self):
        class SeasonIterator(object):

            def __init__(self, values):
                self. values = list(values)
                self.location = 0

            def __iter__(self):
                self.location = 0
                return self

            def __next__(self):
                if self.location >= len(self.values):
                    raise StopIteration
                value = self.values[self.location]
                self.location += 1
                return value
        return SeasonIterator(self.animes)

    def __len__(self):
        return len(self.animes)

    def __hash__(self):
        hash_md5 = hashlib.md5()
        hash_md5.update(str(self.year).encode())
        hash_md5.update(self.season_name.encode())
        return int(hash_md5.hexdigest(), 16)

    def __repr__(self):
        return "<{0:s} {1:s} {2:d}>".format(self.__class__.__name__,
                                            self.season_name, self.year)



__authors__ = ""
__copyright__ = "(c) 2014, pymal"
__license__ = "BSD License"
__contact__ = "Name Of Current Guardian of this file <email@address>"


def load(function):
    """
    This decorator checking of the class was loaded and load it if needed.
    For lazy.
    Needs attribute _is_loaded and a function reload().
    """
    def _load_wrapper(self, *args):
        if not self._is_loaded:
            self.reload()
        return function(self, *args)
    return _load_wrapper


def my_load(function):
    """
    This decorator checking of the class was loaded and load it if needed.
    For lazy.
    Needs attribute _is_my_loaded and a function my_reload().
    """
    def _my_load_wrapper(self, *args):
        if not self._is_my_loaded:
            self.my_reload()
        return function(self, *args)
    return _my_load_wrapper


class Singleton(type):
    """
    Singleton.
    
    Returns:
        An existsing instance.
    """
    _instances = dict()

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class SingletonFactory(type):
    """
    Singleton Factory - keeps one object with the same hash
        of the same cls.
    
    Returns:
        An existing instance.
    """
    __instances = dict()

    def __call__(cls, *args, **kwargs):
        if cls not in cls.__instances:
            cls.__instances[cls] = dict()
        new_obj = super(SingletonFactory, cls).__call__(*args, **kwargs)
        if hash(new_obj) not in cls.__instances[cls]:
            cls.__instances[cls][hash(new_obj)] = new_obj
        return cls.__instances[cls][hash(new_obj)]

    def _unregiter(cls, obj):
        cls.__instances[cls].pop(hash(obj))



__authors__ = ""
__copyright__ = "(c) 2014, pymal"
__license__ = "BSD License"
__contact__ = "Name Of Current Guardian of this file <email@address>"

import hashlib
from xml.etree import ElementTree
from urllib import parse, request

import bs4
from requests.auth import HTTPBasicAuth

from pymal import global_functions
from pymal import decorators
from pymal import exceptions
from pymal import consts
from pymal import AccountAnimes
from pymal import AccountMangas

__all__ = ['Account']


class Account(object, metaclass=decorators.SingletonFactory):
    """
    """
    __all__ = ['animes', 'mangas', 'reload', 'search', 'auth_connect',
               'connect', 'is_user_by_name', 'is_user_by_id', 'is_auth']
    
    __AUTH_CHECKER_URL =\
        request.urljoin(consts.HOST_NAME, r'api/account/verify_credentials.xml')
    __SEARCH_URL = request.urljoin(consts.HOST_NAME, 'api/{0:s}/search.xml')
    __ANIME_SEARCH_URL = __SEARCH_URL.format('anime')
    __MANGA_SEARCH_URL = __SEARCH_URL.format('manga')

    __FRIENDS_URL = request.urljoin(consts.HOST_NAME, 'profile/{0:s}/friends')

    __MY_LOGIN_URL = request.urljoin(consts.HOST_NAME, 'login.php')
    __DATA_FORM = 'username={0:s}&password={1:s}&cookie=1&sublogin=Login'

    def __init__(self, username: str, password: str or None=None):
        """
        """
        self.__username = username
        self.__password = password
        self.connect = global_functions.connect
        self.__user_id = 0
        self.__auth_object = None
        self.__cookies = dict()

        self.__animes = None
        self.__mangas = None
        self.__friends_url = self.__FRIENDS_URL.format(self.username)
        self.__friends = None

        if password is not None:
            self.change_password(password)

    @property
    def username(self) -> str:
        return self.__username

    @property
    def user_id(self) -> int:
        if not self.is_auth:
            raise exceptions.UnauthenticatedAccountError(self.username)
        return self.__user_id

    @property
    def mangas(self) -> AccountMangas.AccountMangas:
        if self.__mangas is None:
            self.__mangas = AccountMangas.AccountMangas(self.username, self)
        return self.__mangas

    @property
    def animes(self) -> AccountAnimes.AccountAnimes:
        if self.__animes is None:
            self.__animes = AccountAnimes.AccountAnimes(self.username, self)
        return self.__animes

    @property
    def friends(self) -> set:
        class FriendsFrozenSet(set):
            def __init__(self, account: Account):
                super().__init__()

                self.account = account
                self.reload()

            def reload(self):
                self.clear()
                div_wrapper = global_functions.get_content_wrapper_div(self.account._Account__friends_url, self.account.connect)
                assert div_wrapper is not None

                list_div_friend = div_wrapper.findAll(name="div", attrs={"class": "friendBlock"})
                for div_friend in list_div_friend:
                    div_pic = div_friend.find(name="div", attrs={'class': 'picSurround'})
                    assert div_pic is not None

                    splited_friend_url = div_pic.a['href'].split('/profile/', 1)
                    assert len(splited_friend_url) == 2

                    self.add(Account(splited_friend_url[1]))

        self.__friends = FriendsFrozenSet(account=self)
        return self.__friends

    def search(self, search_line: str, is_anime: bool=True) -> set:
        """
        """
        params = {'q': search_line}
        if is_anime:
            base_url = self.__ANIME_SEARCH_URL
            from pymal import Anime
            searched_object = Anime.Anime
            account_object_list = self.animes
        else:
            base_url = self.__MANGA_SEARCH_URL
            from pymal import Manga
            searched_object = Manga.Manga
            account_object_list = self.mangas

        url_parts = list(parse.urlparse(base_url))
        query = dict(parse.parse_qsl(url_parts[4]))
        query.update(params)
        url_parts[4] = parse.urlencode(query)
        search_url = parse.urlunparse(url_parts)

        data = self.auth_connect(search_url)
        entries = bs4.BeautifulSoup(data).body.anime.findAll(
            name='entry', recursive=False)

        def get_object(entry):
            object_id = int(entry.id.text)
            if object_id in account_object_list:
                return list(filter(
                    lambda x: x == object_id,
                    account_object_list
                ))[0]
            return searched_object(object_id)
        return set(map(get_object, entries))

    def change_password(self, password: str) -> bool:
        """
        Checking if the new password is valid
        """
        self.__auth_object = HTTPBasicAuth(self.username, password)
        data = self.auth_connect(self.__AUTH_CHECKER_URL)
        if data == 'Invalid credentials':
            self.__auth_object = None
            self.__password = None
            return False
        xml_user = ElementTree.fromstring(data)

        assert 'user' == xml_user.tag, 'user == {0:s}'.format(xml_user.tag)
        l = list(xml_user)
        xml_username = l[1]
        assert 'username' == xml_username.tag,\
            'username == {0:s}'.format(xml_username.tag)
        assert self.username == xml_username.text.strip(),\
            'username = {0:s}'.format(xml_username.text.strip())

        xml_id = l[0]
        assert 'id' == xml_id.tag, 'id == {0:s}'.format(xml_id.tag)
        self.__user_id = int(xml_id.text)

        self.__password = password

        data_form = self.__DATA_FORM.format(self.username, password).encode('utf-8')
        self.connect(self.__MY_LOGIN_URL, data=data_form)

        return True

    def auth_connect(self, url: str, data: str or None=None,
                     headers: dict or None=None) -> str:
        """
        """
        assert self.is_auth, "Not auth yet!"
        return global_functions._connect(url, data=data, headers=headers,
                                         auth=self.__auth_object).text.strip()

    @property
    def __cookies_string(self) -> str:
        return ";".join(["=".join(item)for item in self.__cookies.items()])

    @property
    def is_auth(self) -> bool:
        """
        """
        return self.__auth_object is not None

    def __repr__(self):
        return "<Account username: {0:s}>".format(self.username)

    def __hash__(self):
        hash_md5 = hashlib.md5()
        hash_md5.update(self.username.encode())
        return int(hash_md5.hexdigest(), 16)



__authors__ = ""
__copyright__ = "(c) 2014, pymal"
__license__ = "BSD License"
__contact__ = "Name Of Current Guardian of this file <email@address>"

__all__ = ['Account', 'Anime', 'Manga', 'Seasons']


def get_version():
    return '0.5b2'



__authors__ = ""
__copyright__ = "(c) 2014, pymal"
__license__ = "BSD License"
__contact__ = "Name Of Current Guardian of this file <email@address>"

__all__ = ['USER_AGENT', 'HOST_NAME', 'DEBUG',
           'RETRY_NUMBER', 'RETRY_SLEEP', 'LONG_SITE_FORMAT_TIME',
           'SHORT_SITE_FORMAT_TIME', 'LONG_SITE_FORMAT_TIME',
           'MALAPPINFO_FORMAT_TIME', 'MALAPPINFO_NONE_TIME', 'MALAPI_FORMAT_TIME',
           'MALAPI_NONE_TIME']


USER_AGENT = 'api-indiv-0829BA2B33942A4A5E6338FE05EFB8A1'

HOST_NAME = "http://myanimelist.net"

DEBUG = True
RETRY_NUMBER = 4
RETRY_SLEEP = 1

SHORT_SITE_FORMAT_TIME = '%b %Y'
LONG_SITE_FORMAT_TIME = '%b %d, %Y'
MALAPPINFO_FORMAT_TIME = "%Y-%m-%d"
MALAPPINFO_NONE_TIME = "0000-00-00"
MALAPI_FORMAT_TIME = "%Y%m%d"
MALAPI_NONE_TIME = "00000000"



from setuptools import setup


# Dynamically calculate the version based on pymal.VERSION.
version = __import__('pymal').get_version()


setup(
    name='pymal',
    packages=['pymal'],
    version=version,
    description='A python api for the website MyAnimeList (or MAL).',
    author='pymal-developers',
    license="BSD",
    url='https://bitbucket.org/pymal-developers/pymal/',
    keywords=["MyAnimeList", "MAL", "pymal"],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Natural Language :: Japanese',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.0',
        'Programming Language :: Python :: 3.1',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Topic :: Database',
        'Topic :: Database :: Front-Ends',
        'Topic :: Home Automation',
        'Topic :: Internet',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Site Management',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: User Interfaces',
    ],
    install_requires=[
        'requests>=2.3.0',
        'beautifulsoup4>=4.3.2',
        'httpcache>=0.1.3',
        'html5lib>=0.999',
        'six>=1.1,<1.4',
    ],

)
